const constants = require('./util/constants.json')

const callIfPresent = (func, args, fallback = false) => {
  if(typeof func === 'function') {
    return func(...args)
  } else if(fallback) {
    return fallback(...args)
  }
}

// const pathToKey = (path) => {
//   let key = $fs.utils.resolvePath(path)

//   if(key.startsWith('/a/')) {
//     key.slice(3)
//   }

//   if(key.endsWith('/')) {
//     key = key.slice(0, -1)
//   }

//   return key
// }

const localBasePath = constants.storage.baseKey
const localExt = constants.storage.objectsInFilesystem.ext

console.log('LOCAL BASE PATH', localBasePath)

class EmulatedIDB {
  constructor() {
    const self = this

    const keyOnlyWrapper = original => key =>
      original(this.getRemoteKey(key) )

    const valueKeyWrapper = original => (item, key) =>
      original(item, this.getRemoteKey(key) )

    this.objectStoreOverrides = {
      get: keyOnlyWrapper,
      delete: keyOnlyWrapper,
      put: valueKeyWrapper,
      add: valueKeyWrapper,
      
      clear: original => () => {
        let out = {}

        this.getLocalRemoteKeys
          .then(keys => Promise.all(
            keys.map(key => localforage.removeItem(key))
          ))
          .catch(err => {
            callIfPresent(out.onerror, [ err ], err => {
              throw err
            })
          })
          .then(() => {
            callIfPresent(out.onsuccess, [ {} ])
          })

        return out
      },

      index: () => targetEntryKey => {
        function openKeyCursor() {
          let out = {}

          self.getLocalRemoteKeys().then(async remoteKeys => {
            let files = await Promise.all(
              remoteKeys
                .map(async remoteKey => {
                  let obj = await localforage.getItem(remoteKey)

                  return {
                    key: obj[targetEntryKey],
                    primaryKey: self.getLocalKey(remoteKey)
                  }
                })
            )

            let index = 0

            const nextItem = () => {
              callIfPresent(out.onsuccess, [
                { target: { result: makeCursor() } }
              ])
            }

            const makeCursor = () => {
              if(index < files.length)
                return Object.assign(files[index], {
                  continue() {
                    index++
                    nextItem()
                  }
                })
              else
                return null
            }

            nextItem()
          })
          
          return out
        }

        return { openKeyCursor }
      }
    }
  }

  createMethodProxy(target, funcs) {
    return new Proxy(target, {
      get(target, prop) {
        console.log('IDB Proxy accessing:', prop)

        if(funcs[prop])
          return funcs[prop](target[prop].bind(target), target)
        else
          return target[prop]
      },

      set(target, prop, value) {
        // Ensure sure handlers like onsuccess and such work properly
        target[prop] = typeof value == 'function' ? value.bind(target) : value
      }
    })
  }

  getRemoteKey(key, includeExt = true) {
    return localBasePath + (includeExt ? key + localExt : key)
  }

  isKeyLocal(key) {
    return key.startsWith(localBasePath) && key.endsWith(localExt)
  }

  getLocalKey(key) {
    return key.slice(localBasePath.length, -localExt.length)
  }

  getLocalRemoteKeys() {
    return localforage.keys()
      .then(keys => 
        keys.filter(key => this.isKeyLocal(key))
      )
  }

  patch(IDBFS) {
    const self = this

    IDBFS.getDB = function getDB(name, callback) {
      const dbEmulator = self.createMethodProxy(top.localforage._dbInfo.db, {
        transaction: original => (name, ...args) => {
          let out = original(['a'], ...args)
  
          if(Array.isArray(name) && name[0] == IDBFS.DB_STORE_NAME)
            out = self.createMethodProxy(out, {
              objectStore: original => () =>
                self.createMethodProxy(original('a'), self.objectStoreOverrides)
            })
  
          // if(Array.isArray(name) )
          //   out = self.createMethodProxy(out, {
          //     objectStore: original => () =>
          //       self.createMethodProxy(original('a'), self.objectStoreOverrides)
          //   })


          return out
        }
      })
  
      callback(null, dbEmulator)
    }
  }

  sync() {
    $explorer.refresh()
  }
}

module.exports = EmulatedIDB